#ifndef _SLINK_H
#define _SLINK_H

#ifdef __KERNEL__

#include <linux/types.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/version.h>
#include <linux/list.h>
#include <linux/parport.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/timex.h>
#endif /* __KERNEL__ */

/* By default SLINK_USE_GET_CYCLE is disabled which forces the
 * driver to use the do_gettimeofday() calls to calculate the
 * pulse timing on the slink bus.  This is by far the most generic
 * solution and hense most portable.  In most situations this is
 * probably what you want, however on slower 486 class machines
 * the timing resolution may be too poor.  If you suspect this
 * may be a problem define this and the driver will use the i386
 * cycle counter to calculate pulse timing.
 */
#undef SLINK_USE_GET_CYCLE

/* Define SLINK_DISABLE_IRQ to disable interrupts during driver
 * writes.  This will prevent the driver from decoding and placing
 * on all the read queues the command just issued.  Instead the
 * command will be manually inserted on all read queues except that
 * of the writer.  
 */
#define SLINK_DISABLE_IRQ    1

/* When SLINK_FRAME_SLEEP is defined the 45ms IR frame length
 * will be padded out as the driver sleeps.  Unfortunately, this
 * only has a 10ms resolution on most i386 systems with HZ=100.
 * This means an IR pulse will be padded to a _minimum_ of the
 * required 45ms.  This is good enough for most devices however
 * there may be some that are more picky.  If SLINK_FRAME_SLEEP is
 * not defined then udelay() will be used instead to get the nessisary 
 * precision.  However, this will cause the kernel to effectively
 * spin and other tasks will not be scheduled until the command 
 * has completed.
 */
#define SLINK_FRAME_SLEEP    1

/* SLINK_CTRL_S_REPEAT is the number of times a Control-S command should 
 * be repeated for the benefit of the attached device.  Since Control-S 
 * is an IR protocol which is prone to many kinds of interference every 
 * command must be repeated several times before a device accepts it.
 * For most devices this threshold appears to be 3 times, but it's possible
 * other devices will need the command repeat more often.
 */
#define SLINK_CTRL_S_REPEAT  3


#define SLINK_NO               8
#define MAX_PACKET_SIZE        32
#define IRQ_BUFFER_SIZE        16384
#define DEBUG_BUFFER_SIZE      1048576

#define SLINK_EXIST            0x01
#define SLINK_BUSY             0x02
#define SLINK_IRQ_FLUSH        0x04
#define SLINK_WRITE_PENDING    0x08
#define SLINK_READ_VERBOSE     0x10
#define SLINK_EXTRA_IRQ        0x20
#define SLINK_OVERFLOW_IRQ     0x40
#define SLINK_UNDERFLOW_IRQ    0x80

#define SLINK_VERBOSE          691
#define SLINK_ERROR_EXTRA_IRQ  692

#ifdef __KERNEL__

#ifndef MODULE_LICENSE
#define MODULE_LICENSE(license)	;
#endif

#undef  MIN
#define MIN(a,b)  (((a) <= (b)) ? (a) : (b))

#ifdef SLINK_DEVICE_IDS
/* Device ID codes to be used when calling the driver appear in the
 * first row of the following table.  This is abstraction is done for
 * three reasons.
 *   1) This allows a more logical grouping of devices in to category's
 *      when accessing the driver.  For instaces all CD players start 0x9?.
 *   2) More importantly it allows us to handle 5, 8 and 13 bit device
 *      ID codes without forcing that complexity on the caller.
 *   3) We can add new devices without changing the device ID users
 *      are accustomed to using.  This is good for compatibility.
 *
 * NOTE: This is NOT and exhaustive list of device ID codes and most
 *       of these are not even confirmed to be correct.  Additions and
 *       corrections to this list would be appreciated!
 */
#define DRIVER_ID    0
#define CTRL_A1_ID   1
#define CTRL_S_BITS  2
#define CTRL_S_ID    3

int slink_device_ids[][4] = {
  { 0x00,  0x00, 5,  0x00   },   /* Unknown   */
  { 0x10,  0x00, 8,  0x2A   },   /* Projector */
  { 0x20,  0x00, 5,  0x0E   },   /* Tape   */
  { 0x30,  0x00, 5,  0x07   },   /* DAT    */
  { 0x40,  0x00, 8,  0xED   },   /* DSS    */
  { 0x50,  0x00, 5,  0x10   },   /* TV     */
  { 0x60,  0x00, 5,  0x0C   },   /* LD-1   */
  { 0x70,  0x00, 5,  0x1E   },   /* MD-1   */
  { 0x80,  0x00, 5,  0x08   },   /* VCR-1  */
  { 0x81,  0x00, 5,  0x16   },   /* VCR-2  */
  { 0x82,  0x00, 5,  0x1A   },   /* VCR-3  */
  { 0x83,  0x00, 8,  0x2D   },   /* VCR-1 (alternate) */
  { 0x84,  0x00, 8,  0xED   },   /* VCR-2 (alternate) */
  { 0x85,  0x00, 8,  0xFD   },   /* VCR-3 (alternate) */
  { 0x90,  0x90, 5,  0x11   },   /* CD-1   */
  { 0x91,  0x91, 8,  0x9C   },   /* CD-2   */
  { 0x92,  0x92, 8,  0x8A   },   /* CD-3   */
  { 0x93,  0x93, 0,  0x00   },   /* CD-1 (alternate) */
  { 0x94,  0x94, 0,  0x00   },   /* CD-2 (alternate) */
  { 0x95,  0x95, 0,  0x00   },   /* CD-3 (alternate) */
  { 0xA0,  0x00, 13, 0x0B92 },   /* DVD-1  */
  { 0xA1,  0x00, 13, 0x0B44 },   /* DVD-2  */
  { 0xA2,  0x00, 13, 0x0B5F },   /* DVD-3  */
  { 0xAA,  0x00, 13, 0x0B46 },   /* DVD-1 MegaChanger Commands */
  { 0xB0,  0x00, 5,  0x16   },   /* Tuner  */
  { 0xC0,  0xC0, 5,  0x06   },   /* Amp    */
  { 0xC1,  0xC1, 5,  0x06   },   /* Amp    */
  { 0xC3,  0xC3, 5,  0x06   },   /* Amp    */
  { -1,      -1, -1, -1     }
}; /* slink_device_ids */
#endif


/* Macros to conditionally disable or enable interrupts based on the
 * compile time option.  Originally I wrapped each instance of this in
 * the source but that began to get cluttered.
 */
#ifdef SLINK_DISABLE_IRQ 
  #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
    #define DISABLE_IRQ(flags)   save_flags(flags); cli();
    #define ENABLE_IRQ(flags)    restore_flags(flags);
  #else
    extern spinlock_t slink_lock;
    #define DISABLE_IRQ(flags)   spin_lock_irqsave(&slink_lock, flags);
    #define ENABLE_IRQ(flags)    spin_unlock_irqrestore(&slink_lock, flags);
  #endif
#else
  #define DISABLE_IRQ(flags)   ;
  #define ENABLE_IRQ(flags)    ;
#endif /* SLINK_DISABLE_IRQ */


/* ----------------------- Debug --------------------------- */

#define D_WARN			0x01
#define D_ERROR			0x02
#define D_INFORM		0x04
#define D_MEMORY		0x08
#define D_DEBUG			0x10
#define D_TRACE			0x20
#define D_PROC			0x40

#if defined(x86_64)
#define FMT_SIZE_T		"%ld"
#else
#define FMT_SIZE_T              "%d"
#endif
    
/* Only D_WARN|D_ERROR|D_INFORM can go to the console, other debugging
 * will go to /proc/slink_debug because of its verbosity.
 */
#define PRINT(type, fmt, args...)                                         \
do {                                                                      \
  if ((slink_debug & 0x07) & (type))                                      \
    printk(KERN_ERR "S-Link: " fmt , ## args);                            \
                                                                          \
  if ((slink_debug & D_PROC) && (debug_buffer.buf != NULL)) {             \
    int len, chunk;                                                       \
    struct timeval now;							  \
    do_gettimeofday(&now);						  \
                                                                          \
    len = sprintf(debug_buffer.tmp,                                       \
      "0x%02X: %ld.%06ld:%d:%-24s" fmt, type, now.tv_sec, now.tv_usec,    \
      __LINE__, __FUNCTION__ , ## args);                                  \
    if (debug_buffer.head + len < debug_buffer.buf + DEBUG_BUFFER_SIZE) { \
      memcpy(debug_buffer.head, debug_buffer.tmp, len);                   \
      debug_buffer.head += len;                                           \
    } else {                                                              \
      chunk = debug_buffer.buf + DEBUG_BUFFER_SIZE - debug_buffer.head;   \
      memcpy(debug_buffer.head, debug_buffer.tmp, chunk);                 \
      memcpy(debug_buffer.buf, debug_buffer.tmp + chunk, len - chunk);    \
      debug_buffer.head = debug_buffer.buf + (len - chunk);               \
    }                                                                     \
                                                                          \
    if (debug_buffer.size < DEBUG_BUFFER_SIZE) {                          \
      if (debug_buffer.size + len < DEBUG_BUFFER_SIZE)                    \
        debug_buffer.size += len;                                         \
      else                                                                \
        debug_buffer.size = DEBUG_BUFFER_SIZE;                            \
    }                                                                     \
  }                                                                       \
} while (0)

#define SPRINT(addr, args...)                                             \
do {                                                                      \
  if (len <= max) {                                                       \
    len += sprintf((addr) , ## args);                                     \
  } else {                                                                \
    up(&(slink_table[i].user_sem));                                       \
    *eof = 1;                                                             \
    return len;                                                           \
  }                                                                       \
} while (0)

#define PRINT_COMMAND(buf, msg)						  \
  buf += sprintf(buf, "%-16s|", msg)
#define TRACE(fmt, args...)						  \
  PRINT(D_TRACE, "%s(" fmt ")\n", __FUNCTION__, ## args)


/* ------------------- Memory Handling --------------------- */

extern unsigned long slink_memory;

#define slink_memory_inc(ptr, size)	slink_memory += (size);
#define slink_memory_dec(ptr, size)	slink_memory -= (size);

#define KMALLOC(ptr, size)						\
do {									\
  long s = (size);							\
  (ptr) = kmalloc(s, GFP_KERNEL);					\
  if ((ptr) == NULL) {							\
    PRINT(D_MEMORY, "Out of memory, unable to allocate %ld\n",s);	\
  } else {								\
    slink_memory_inc((ptr), s);						\
    memset((ptr), 0, s);						\
  }									\
  PRINT(D_MEMORY, "KMalloc'ed %ld at %p (total %ld)\n",			\
        s, (ptr), slink_memory);					\
} while (0)

#define KFREE(ptr, size)						\
do {									\
  long s = (size);							\
  if ((ptr) == NULL) {							\
    PRINT(D_MEMORY, "Attempted to free %ld from NULL pointer\n", s);	\
    break;								\
  }									\
  kfree(ptr);								\
  slink_memory_dec((ptr), s);						\
  PRINT(D_MEMORY, "KFree'd %ld at %p (total %ld)\n",			\
        s, (ptr), slink_memory);					\
} while (0)

   
/* ---------------------- Structures ------------------------ */

struct slink_cycles {
  cycles_t init;
  cycles_t zero;
  cycles_t one;
  cycles_t clear;
  cycles_t tolerance;
  cycles_t ir;
}; /* slink_cycles() */


struct slink_db {
  char tmp[256];
  char *buf;
  char *head;
  int size;
}; /* slink_db */


struct slink_user {
  struct list_head list;
  struct list_head in_queue;
  unsigned long in_queue_depth;
  unsigned long flags;
  int id;
}; /* slink_user */


struct slink_packet {
  struct list_head list;
  char data[MAX_PACKET_SIZE];
  unsigned long size, flags, transmitted;
  cycles_t calc_time, real_time;
}; /* struct slink_packet */


struct slink_buffer {
  void *buffer;
  void *head, *tail;
  cycles_t last;
  unsigned long size;
}; /* struct slink_buffer */


struct slink_struct {
  struct pardevice *dev;
  unsigned long flags;
  unsigned long users;
  unsigned long in_packets;
  unsigned long out_packets;
  unsigned long corrupted_packets;
  unsigned long corrected_packets;
  unsigned long collision_packets;
  unsigned long interrupts;
  unsigned long out_queue_depth;

  struct slink_buffer irqs;
  struct tasklet_struct write_tasklet;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  struct tasklet_struct read_tasklet;
#else
  struct work_struct read_work;
#endif
  wait_queue_head_t read_waitq;
  wait_queue_head_t write_waitq;

  struct list_head user_queue;
  struct list_head out_queue; 

  struct semaphore user_sem;
  struct semaphore read_sem;
}; /* struct slink_struct */


/* ---------------------- Variables ------------------------ */

extern unsigned int slink_debug;
extern unsigned int slink_major;
extern struct slink_struct slink_table[SLINK_NO];
extern struct slink_cycles slink_time;
extern struct slink_db debug_buffer;


/* ------------------------ General ------------------------- */

extern int slink_print_packet(char *, struct slink_packet *);
extern int slink_read_proc(char *, char **, off_t, int, int *, void *);
extern int slink_debug_read_proc(char *, char **, off_t, int, int *, void *);

#ifdef SLINK_USE_GET_CYCLES

#define QUERY_INTERVAL(value, time)	\
  value = (get_cycles() - (time))

#define SET_INTERVAL(value, time)	\
do {					\
  if (time == 0) {			\
    value = get_cycles();		\
  } else {				\
    cycles_t start = get_cycles();	\
    udelay(time);			\
    value = get_cycles() - start;	\
  }					\
} while(0)

#else

#define QUERY_INTERVAL(value, time)					\
do {									\
  struct timeval start;							\
  do_gettimeofday(&start);						\
  value = (cycles_t)(start.tv_sec * 1000000 + start.tv_usec) - time;	\
} while(0)

#define SET_INTERVAL(value, time)					\
do {									\
  struct timeval start;							\
									\
  if (time == 0) {							\
    do_gettimeofday(&start);						\
    value = (cycles_t)(start.tv_sec * 1000000 + start.tv_usec);		\
  } else {								\
    value = (cycles_t)(time);						\
  }									\
} while(0)
#endif /* SLINK_USE_GET_CYCLES */

#define NEXT_IRQ(ptr, type)	((ptr->type < (ptr->buffer + 		\
                                 IRQ_BUFFER_SIZE - sizeof(cycles_t))) ? \
				 (ptr->type += sizeof(cycles_t)) :	\
				 (ptr->type  = ptr->buffer))
#define NEXT_IRQ_HEAD(ptr)						\
do {									\
  NEXT_IRQ(((struct slink_buffer *)ptr), head);				\
  ((struct slink_buffer *)ptr)->size += sizeof(cycles_t);		\
} while(0)

#define NEXT_IRQ_TAIL(ptr)						\
do {									\
  NEXT_IRQ(((struct slink_buffer *)ptr), tail);				\
  ((struct slink_buffer *)ptr)->size -= sizeof(cycles_t);		\
} while(0)

#define THETA             600
#define PORT_CLEAR        4300   /* us */
#define PORT_TOLERANCE    300    /* us */

#define INIT_LENGTH       (5 * THETA)   /* us */
#define ONE_LENGTH        (3 * THETA)   /* us */
#define ZERO_LENGTH       (2 * THETA)   /* us */
#define IR_LENGTH         45000         /* us */

#define LO(mask)          (outb_p(inb_p(slink_base) & ~mask, slink_base))
#define HI(mask)          (outb_p(inb_p(slink_base) | mask, slink_base))
#define SLEEP             (udelay(THETA))

/* Collisions can only be detected when the control-A1 bus should be
 * floating at 5v (LO) and is instead grounded.  Collisions cannot occur
 * over Control-S because it does not using a bus scheme and the natural
 * line state is to ground.
 */
static inline int collision(int slink_base, unsigned char mask) {

  /* Collisions can only occur on Control-A1 bus */
  if (mask != 0x01)
    return 0;

  /* Input pin 11 is tied to output pin 2, and when we write a 0 to pin 2
   * the slink bus will float high at 5v.  If some other device is trying 
   * to send they will ground to bus to 0v.  So when we read back pin 11
   * we should get 0 if the line is floating high as expected, or 1 if
   * some other device is trying to send in which case a collision has
   * occured.  Remember the Control-A1 hardware is the inverse of the
   * Control-S hardware.
   */
  if (inb_p((slink_base + 1) & 0x80) == 0x80)
    return 1;

  return 0;
} /* collision() */

/* init()/zero()/one() will return time taken to send (>0), 
 * or a 0 will be returned if a collision occured. 
 */
static inline cycles_t init(int slink_base, unsigned char mask) {
  HI(mask); SLEEP; SLEEP; SLEEP; SLEEP;
  LO(mask);

  if (collision(slink_base, mask) == 1)
    return 0;

  SLEEP;

  if (collision(slink_base, mask) == 1)
    return 0;

  return slink_time.init;
} /* init() */


static inline cycles_t one(int slink_base, unsigned char mask) {
  HI(mask); SLEEP; SLEEP;
  LO(mask);

  if (collision(slink_base, mask) == 1)
    return 0;

  SLEEP;

  if (collision(slink_base, mask) == 1)
    return 0;

  return slink_time.one;
} /* one() */


static inline cycles_t zero(int slink_base, unsigned char mask) {
  HI(mask); SLEEP;
  LO(mask); 

  if (collision(slink_base, mask) == 1)
    return 0;

  SLEEP;

  if (collision(slink_base, mask) == 1)
    return 0;

  return slink_time.zero;
} /* zero() */

#endif /* __KERNEL__ */

#endif /* _SLINK_H */
