/* Sony S-Link Protocol Device Driver for 2.4.x Linux Kernels
 * Copyright (c) 2002 Brian Behlendorf, behlendorf1@llnl.gov
 *
 * The S-Link driver is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Full text for the GNU General Public License is available at
 * http://www.gnu.org/copyleft/gpl.html or by writing to the Free Software 
 * Foundation, Inc., 59 Temple Place Suite 330, Boston, MA  02111-1307,USA.
 */

#define SLINK_DEVICE_IDS
#include "slink.h"
#include "config.h"

static unsigned int slink_size = 0;
struct slink_struct slink_table[SLINK_NO];
struct timer_list slink_ctrl_s_timer;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
spinlock_t slink_lock;
#endif

unsigned int slink_major;
unsigned int slink_debug = (D_ERROR | D_INFORM);
unsigned int slink_probe = 1;
unsigned long slink_memory;
struct slink_cycles slink_time; 
struct slink_db debug_buffer;


/* ------------- Module Prototypes/Callbacks  ----------------- */

static ssize_t slink_read(struct file *, char *, size_t, loff_t *);
static ssize_t slink_write(struct file *, const char *, size_t, loff_t *);
static void slink_write_finish(struct slink_struct *);
static unsigned int slink_poll(struct file *, poll_table *);
static int slink_ioctl(struct inode *, struct file *,
  unsigned int, unsigned long);
static int slink_open(struct inode *, struct file *);
static int slink_release(struct inode *, struct file *);

static void slink_attach(struct parport *);
static void slink_detach(struct parport *);
static int slink_preempt(void *);
static void slink_kick(void *);
static void slink_irq(int, void *, struct pt_regs *);


static struct file_operations slink_fops = { 
  open:		slink_open,
  release:	slink_release,
  read:		slink_read,
  write:	slink_write,
  poll:		slink_poll,
  ioctl:	slink_ioctl,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
  owner:	THIS_MODULE,
#endif
}; /* struct file_operations */


static struct parport_driver slink_driver = {
  name:		"slink",
  attach:	slink_attach,
  detach:	slink_detach,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  next:		NULL 
#endif
}; /* struct parport_driver */


/* --------------------- Module Tasklets  --------------------- */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
static void slink_read_tasklet(unsigned long private) {
  struct slink_struct *slink_entry = (struct slink_struct *)private;
#else
static void slink_read_work(void *private) {
#endif
  struct slink_struct *slink_entry = private;
  struct slink_packet *packet;
  struct slink_user *user;
  struct list_head *ptr;
  cycles_t t1 = 0, t2 = 0, delta;
  char buffer[MAX_PACKET_SIZE];
  int byte = 0, bit = 7;

  if (down_interruptible(&(slink_entry->read_sem)))
    return;

  QUERY_INTERVAL(delta, slink_entry->irqs.last);

  /* Good read time or read nessisary interrupt buffer getting full */
  if (((delta > slink_time.clear) && (delta < (slink_time.clear * 5))) ||
      (slink_entry->flags & SLINK_IRQ_FLUSH)) {

    while (1) {

      if (slink_entry->irqs.head == slink_entry->irqs.tail)
        break;

      if (byte >= MAX_PACKET_SIZE)
        break;

      t1 = *(cycles_t *)slink_entry->irqs.tail;
      NEXT_IRQ_TAIL(&(slink_entry->irqs));
      t2 = *(cycles_t *)slink_entry->irqs.tail;
      delta = t2 - t1;

      if ((delta > (slink_time.zero - 3 * slink_time.tolerance)) &&
          (delta < (slink_time.zero + slink_time.tolerance)))
        clear_bit(bit, (void *)buffer + byte);
      else
        if ((delta > (slink_time.one - slink_time.tolerance)) &&
            (delta < (slink_time.one + 3 * slink_time.tolerance)))
          set_bit(bit, (void *)buffer + byte);
        else
          if ((byte == 0) && (bit == 7))
            continue;
          else {
            /* Extra short interupt we should just ignore */
            if (delta < (slink_time.zero - 3 * slink_time.tolerance)) {
              slink_entry->corrected_packets++;
              slink_entry->flags |= SLINK_EXTRA_IRQ;
              PRINT(D_WARN, "Extra Interrupt Corrected (%02X %02X-%02X)\n",
                buffer[0] & 0xFF, buffer[1] & 0xFF, buffer[2] & 0xFF);
              continue;
            }

            break;
          }

      if ((bit--) <= 0) {
        byte++;
        bit = 7;
      }
    }

    if ((byte > 0) || (bit < 7)) {	
      if (bit != 7) {
        slink_entry->corrupted_packets++;
        PRINT(D_WARN, "Corrupt Packet (%02X %02X-%02X; Bits %d) "
          "Delta %llu = (%llu - %llu) (size=%ld)\n",
          buffer[0] & 0xFF, buffer[1] & 0xFF, buffer[2] & 0xFF,
          byte * 8 + (7 - bit), delta, t2, t1, slink_entry->irqs.size);
      } else {
        slink_entry->in_packets++;

        if (slink_entry->flags & SLINK_BUSY) {
          for (ptr = slink_entry->user_queue.next;
               ptr != &(slink_entry->user_queue); ptr = ptr->next) {
            user = list_entry(ptr, struct slink_user, list);
 
            KMALLOC(packet, sizeof(struct slink_packet));
            memcpy(packet->data, buffer, byte);
            packet->size  = byte;
            packet->flags = slink_entry->flags;
            list_add_tail(&(packet->list), &(user->in_queue)); 
            user->in_queue_depth++;

            slink_entry->flags &= ~SLINK_EXTRA_IRQ;
            slink_entry->flags &= ~SLINK_OVERFLOW_IRQ;
            slink_entry->flags &= ~SLINK_UNDERFLOW_IRQ;
          }
        }
      }
    }
  }

  if (slink_entry->flags & SLINK_BUSY)
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
    tasklet_schedule(&(slink_entry->read_tasklet));
#else
    schedule_delayed_work(&(slink_entry->read_work), 10);
#endif

  wake_up(&(slink_entry->read_waitq));

  slink_entry->flags &= ~SLINK_IRQ_FLUSH;
  up(&(slink_entry->read_sem));

  return;
} /* slink_read_tasklet() */


/* Send the normal Control-A1, S-Link commands.  On collision 0 is returned */
static cycles_t write_ctrl_a1(int slink_base, struct slink_packet *packet) {
  cycles_t btime, packet_ctime = 0;
  int i = 0, j, dev_id;

  TRACE("%d, %p", slink_base, packet);
  /* Select the Control-A1 device id code, if it's not listerd use 'unknown' */
  dev_id = slink_device_ids[0][CTRL_A1_ID];
  while (slink_device_ids[i][DRIVER_ID] != -1) {
    if ((packet->data[0] & 0xFF) == slink_device_ids[i][DRIVER_ID]) {
      dev_id = slink_device_ids[i][CTRL_A1_ID];
      break;
    }

    i++;
  }

  packet->data[0] = (char)(dev_id & 0xFF);
  if ((btime = init(slink_base, 0x01)) == 0)
    return 0;

  packet_ctime += btime;
  for (i = 0; i < packet->size; i++) {
    for (j = 7; j >= 0; j--) {
      if (test_bit(j, (void *)packet->data + i)) {
        if ((btime = one(slink_base, 0x01)) == 0)
          return 0;

        packet_ctime += btime;
      } else {
        if ((btime = zero(slink_base, 0x01)) == 0)
          return 0;

        packet_ctime += btime;
      }
    }
  }

  packet->transmitted++; 
  return packet_ctime;
} /* write_ctrl_a1() */


/* Write the IR command out  (HI bits -> LO bits) */
static cycles_t write_ir(int slink_base, unsigned char mask, 
    struct slink_packet *packet) {
  cycles_t btime, packet_ctime = 0;
  int i = 0, j, dev_size, dev_id;

  TRACE("%d, 0x%02X, %p", slink_base, mask & 0xFF, packet);
  /* Select the Control-S device id code, if it's not listed use 'unknown' */
  dev_size = slink_device_ids[0][CTRL_S_BITS];
  dev_id   = slink_device_ids[0][CTRL_S_ID];
  while (slink_device_ids[i][DRIVER_ID] != -1) {
    if ((packet->data[0] & 0xFF) == slink_device_ids[i][DRIVER_ID]) {
      dev_size = slink_device_ids[i][CTRL_S_BITS];
      dev_id   = slink_device_ids[i][CTRL_S_ID];
      break;
    }

    i++;
  }

  for (j = 6; j >= 0; j--) {
    if (test_bit(j, (void *)packet->data + 2)) {
      if ((btime = one(slink_base, mask)) == 0)
        return 0;

      packet_ctime += btime;
    } else {
      if ((btime = zero(slink_base, mask)) == 0)
        return 0;

      packet_ctime += btime;
    }
  }

  for (j = (dev_size-1); j >= 0; j--) {
    if (test_bit(j, (void *)&dev_id)) {
      if ((btime = one(slink_base, mask)) == 0)
        return 0;

      packet_ctime += btime;
    } else {
      if ((btime = zero(slink_base, mask)) == 0)
        return 0;

      packet_ctime += btime;
    }
  }

  return packet_ctime;
} /* write_ir() */


static cycles_t write_ctrl_a1_ir(int slink_base, struct slink_packet *packet) {
  cycles_t btime, packet_ctime = 0;
  int i;

  TRACE("%d, %p", slink_base, packet);
  /* Send the INIT-ONE-ONE-ONE-ZERO header */
  if ((btime = init(slink_base, 0x01)) == 0)
    return 0;

  packet_ctime += btime;
  for (i = 0; i < 3; i++) {
    if ((btime = one(slink_base, 0x01)) == 0)
      return 0;

    packet_ctime += btime;
  }

  if ((btime = zero(slink_base, 0x01)) == 0)
    return 0;

  packet_ctime += btime;
  /* Append Control-S cmd  */
  if ((btime = write_ir(slink_base, 0x01, packet)) == 0)
    return 0;

  packet->transmitted++; 
  packet_ctime += btime;
  return packet_ctime;
} /* write_ctrl_a1_ir() */


/* Control-S commands must be repeated atleast 3 times before a device 
 * will react to a IR command.  This is to minimize the effect of EM 
 * noise when the command is delived via a real IR signal.
 */
static cycles_t write_ctrl_s(int slink_base, int port, 
    struct slink_packet *packet) {
  cycles_t btime, packet_ctime = 0;
  int mask;

  TRACE("%d, %d, %p", slink_base, port, packet);
  mask = 0x01 << port;
  if ((btime = init(slink_base, mask)) == 0)
    return 0;

  packet_ctime += btime;
  if ((btime = write_ir(slink_base, mask, packet)) == 0)
    return 0;

  packet->transmitted++; 
  packet_ctime += btime;
  return packet_ctime;
} /* write_ctrl_s() */


static void write_ctrl_s_timer(unsigned long private) {
  struct slink_struct *slink_entry;
  struct slink_packet *packet;
  int slink_base, packet_type;
  unsigned long pad_time, flags = 0;
  cycles_t calc_time;
  
  TRACE("%lu", private);
  slink_entry = (struct slink_struct *)private;
  packet = list_entry(slink_entry->out_queue.next, struct slink_packet, list);
  packet_type = packet->data[1] & 0xFF;
  slink_base = slink_entry->dev->port->base;

  /* After SLINK_CTRL_S_REPEAT sends return control to the user if blocking */ 
  if (packet->transmitted == SLINK_CTRL_S_REPEAT) {
    QUERY_INTERVAL(packet->real_time, packet->real_time);
    packet->calc_time = SLINK_CTRL_S_REPEAT *
      ((IR_LENGTH / (1000000 / HZ) + 1) * (1000000 / HZ));
    slink_write_finish(slink_entry);
    return;
  }

  /* Measure the real time from the start of the first pulse */
  if (packet->transmitted == 0)
    SET_INTERVAL(packet->real_time, 0);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  parport_get_port(slink_entry->dev->port);
#endif
  DISABLE_IRQ(flags);
    
  calc_time = write_ctrl_s(slink_base, packet_type - 0xE0, packet);
  pad_time  = jiffies + ((unsigned)(IR_LENGTH * HZ) / 1000000) + 1;
  
  ENABLE_IRQ(flags);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  parport_put_port(slink_entry->dev->port);
#endif

  init_timer(&slink_ctrl_s_timer);
  slink_ctrl_s_timer.function = write_ctrl_s_timer;
  slink_ctrl_s_timer.data     = private;
  slink_ctrl_s_timer.expires  = pad_time;
  add_timer(&slink_ctrl_s_timer);

  return;
} /* write_ctrl_s_timer() */


static void slink_write_tasklet(unsigned long private) {
  struct slink_struct *slink_entry = (struct slink_struct *)private;
  struct slink_packet *packet;
  int slink_base, packet_type;
  unsigned long flags = 0;
  cycles_t delta;
#ifndef SLINK_FRAME_SLEEP
  int i;
#endif
  
  TRACE("%lu", private);
  /* Reschedule until it appears the port is clear */
  QUERY_INTERVAL(delta, slink_entry->irqs.last);
  if (delta < slink_time.clear) {
    tasklet_schedule(&(slink_entry->write_tasklet));
    return;
  }

  /* Nothing to write, signal the write() call to complete if blocking */
  if (list_empty(slink_entry->out_queue.next)) {
    slink_entry->flags &= ~SLINK_WRITE_PENDING;
    wake_up(&(slink_entry->write_waitq));
    return;
  }

  packet = list_entry(slink_entry->out_queue.next, struct slink_packet, list);
  packet_type = packet->data[1] & 0xFF;
  packet->transmitted = 0;
  slink_base = slink_entry->dev->port->base;

  if ((packet_type >= 0xE1) && (packet_type <= 0xE7)) {  /* Control-S */
#ifdef SLINK_FRAME_SLEEP
    init_timer(&slink_ctrl_s_timer);
    slink_ctrl_s_timer.function = write_ctrl_s_timer;
    slink_ctrl_s_timer.data     = private;
    slink_ctrl_s_timer.expires  = jiffies + 1;
    add_timer(&slink_ctrl_s_timer);
    return;

#else /* SLINK_FRAME_SLEEP */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
    parport_get_port(slink_entry->dev->port);
#endif
    DISABLE_IRQ(flags);
    SET_INTERVAL(packet->real_time, 0);
    
    for (i = 0; i < SLINK_CTRL_S_REPEAT; i++) {
      packet->calc_time = write_ctrl_s(slink_base, packet_type - 0xE0, packet);
      udelay(IR_LENGTH - packet->calc_time);
    }
#endif /* SLINK_FRAME_SLEEP */

    packet->calc_time = IR_LENGTH * SLINK_CTRL_S_REPEAT;
  } else {
    /* These commands will be sent NOW */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
    parport_get_port(slink_entry->dev->port);
#endif
    DISABLE_IRQ(flags);
    SET_INTERVAL(packet->real_time, 0);

    if (packet_type == 0xE0) {  /* Control-A1 embedded IR */
      packet->calc_time = write_ctrl_a1_ir(slink_base, packet);
    } else {                    /* Control-A1 */
      packet->calc_time = write_ctrl_a1(slink_base, packet);
    }
  }
  
  QUERY_INTERVAL(packet->real_time, packet->real_time);
  ENABLE_IRQ(flags);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  parport_put_port(slink_entry->dev->port);
#endif

  slink_write_finish(slink_entry);
  
  return;
} /* slink_write_tasklet() */


static void slink_write_finish(struct slink_struct *slink_entry) {
  struct slink_packet *packet;

  TRACE("%p", slink_entry);
  packet = list_entry(slink_entry->out_queue.next, struct slink_packet, list);

  if (packet->calc_time == 0) {
    slink_entry->collision_packets++;
    PRINT(D_WARN, "Packet collision, RETRY");
  } else {
    if ((((int)packet->real_time - (int)packet->calc_time) <
        -((int)packet->calc_time / 50)) ||
        (((int)packet->real_time - (int)packet->calc_time) >
         ((int)packet->calc_time / 50)))
      PRINT(D_WARN, "Packet Write Error; Send Time %lld, Expected Time %lld\n",
        packet->real_time, packet->calc_time);

    list_del(slink_entry->out_queue.next);
    KFREE(packet, sizeof(struct slink_packet));
    slink_entry->out_queue_depth--;
  }

  if (!list_empty(slink_entry->out_queue.next)) {
    tasklet_schedule(&(slink_entry->write_tasklet));
  } else {
    wake_up(&(slink_entry->write_waitq));
    slink_entry->flags &= ~SLINK_WRITE_PENDING;
  }

  return;
} /* slink_write_finish() */


/* ------------------ Module File Ops  --------------------- */

static ssize_t slink_read(struct file *filp, char *buf,
    size_t count, loff_t *ppos) {
  unsigned int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
  struct slink_user *user = (struct slink_user *)filp->private_data;
  struct slink_packet *packet;
  ssize_t size;

  TRACE("%p, %p, " FMT_SIZE_T ", %p", filp, buf, count, ppos);
  if (down_interruptible(&(slink_table[minor].user_sem)))
    return -EINTR;

  if (!(filp->f_flags & O_NONBLOCK)) {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
    tasklet_schedule(&(slink_table[minor].read_tasklet));
    interruptible_sleep_on(&(slink_table[minor].read_waitq));
#else
    schedule_delayed_work(&(slink_table[minor].read_work), 10);
    if (wait_event_interruptible(slink_table[minor].read_waitq, 
        !list_empty(user->in_queue.next))) {
      up(&(slink_table[minor].user_sem));
      return -EINTR;
    }
#endif
  }

  if (list_empty(user->in_queue.next)) {
    up(&(slink_table[minor].user_sem));
    return -ENODATA;
  }

  if ((packet=list_entry(user->in_queue.next,struct slink_packet,list))==NULL){
    up(&(slink_table[minor].user_sem));
    return -ENODATA;
  }

  /* User wishes an error on packets with extra IRQs */
  if ((user->flags & SLINK_EXTRA_IRQ) && (packet->flags & SLINK_EXTRA_IRQ)) {
    size = -EIO;
    goto read_error;
  }

  /* User wishes a ascii decoded version of the packet */
  if (user->flags & SLINK_READ_VERBOSE) {
    char text[256];

    if ((size = (slink_print_packet(text, packet) + 1)) > count)
      size = count;

    if (copy_to_user(buf, text, size)) {
      up(&(slink_table[minor].user_sem));
      return -EFAULT;
    }
  } else {
    size = packet->size;

    if (copy_to_user(buf, packet->data, packet->size)) {
      up(&(slink_table[minor].user_sem));
      return -EFAULT;
    }
  }

read_error:
  list_del(user->in_queue.next);
  user->in_queue_depth--;
  KFREE(packet, sizeof(struct slink_packet));

  up(&(slink_table[minor].user_sem));

  return size; 
} /* slink_read() */


static ssize_t slink_write(struct file *filp, const char *buf,
    size_t count, loff_t *ppos) {
  unsigned int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
  struct slink_packet *packet1;
#ifdef SLINK_DISABLE_IRQ
  struct slink_user *user2, *user1 = (struct slink_user *)filp->private_data;
  struct slink_packet *packet2;
  struct list_head *ptr;
#endif /* SLINK_DISABLE_IRQ */

  TRACE("%p, %p, " FMT_SIZE_T ", %p", filp, buf, count, ppos);
  if (count > MAX_PACKET_SIZE)
    return -EOVERFLOW;

  if (down_interruptible(&(slink_table[minor].user_sem)))
    return -EINTR;

  KMALLOC(packet1, sizeof(struct slink_packet));

  if (packet1 == NULL) {
    up(&(slink_table[minor].user_sem));
    return -ENOMEM;
  }

  if (copy_from_user(packet1->data, buf, count)) {
    KFREE(packet1, sizeof(struct slink_packet));
    up(&(slink_table[minor].user_sem));
    return -EFAULT;
  }

  packet1->size = count;
  slink_table[minor].out_packets++;
  slink_table[minor].out_queue_depth++;
  list_add_tail(&(packet1->list), &(slink_table[minor].out_queue)); 

#ifdef SLINK_DISABLE_IRQ
  /* Add the packet to everyone elses read queue */
  for (ptr = slink_table[minor].user_queue.next;
       ptr != &(slink_table[minor].user_queue); ptr = ptr->next) {
    user2 = list_entry(ptr, struct slink_user, list);

    if (user1->id == user2->id)
      continue;

    KMALLOC(packet2, sizeof(struct slink_packet));
    memcpy(packet2->data, packet1->data, packet1->size);
    packet2->size = packet1->size;

    list_add_tail(&(packet2->list), &(user2->in_queue)); 
    user2->in_queue_depth++;

  }
#endif /* SLINK_DISABLE_IRQ */

  if (!(slink_table[minor].flags & SLINK_WRITE_PENDING)) {
    slink_table[minor].flags |= SLINK_WRITE_PENDING;
    tasklet_schedule(&(slink_table[minor].write_tasklet));
  }

  /* If we're blocking, wait for the write to complete */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
  if (!(filp->f_flags & O_NONBLOCK)) {
    interruptible_sleep_on(&(slink_table[minor].write_waitq));
#else
  if (!(filp->f_flags & O_NONBLOCK)) {
    if (wait_event_interruptible(slink_table[minor].write_waitq, 
        list_empty(slink_table[minor].out_queue.next))) {
      up(&(slink_table[minor].user_sem));
      return -EINTR;
    }
  }
#endif

  up(&(slink_table[minor].user_sem));

  return count;
} /* slink_write() */


static unsigned int slink_poll(struct file *filp, poll_table *table) {
  unsigned int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
  struct slink_user *user = (struct slink_user *)filp->private_data;
  unsigned mask = 0;
  cycles_t delta;

  poll_wait(filp, &(slink_table[minor].read_waitq), table);
  poll_wait(filp, &(slink_table[minor].write_waitq), table);
  QUERY_INTERVAL(delta, slink_table[minor].irqs.last);
  
  if (!list_empty(user->in_queue.next))
    mask |= POLLIN | POLLRDNORM;

  if (delta > slink_time.clear)
    mask |= POLLOUT | POLLWRNORM;

  return mask;
} /* slink_poll() */


static int slink_ioctl(struct inode *inode, struct file *filp,
    unsigned int cmd, unsigned long arg) {
  struct slink_user *user = (struct slink_user *)filp->private_data;

  TRACE("%p, %p, %u, %lu", inode, filp, cmd, arg);
  switch (cmd) {
    case SLINK_VERBOSE:
      user->flags = (arg ? (user->flags | SLINK_READ_VERBOSE) :
        user->flags & ~SLINK_READ_VERBOSE);
      return 0;
    case SLINK_ERROR_EXTRA_IRQ:
      user->flags = (arg ? (user->flags | SLINK_EXTRA_IRQ) :
        user->flags & ~SLINK_EXTRA_IRQ);
      return 0;
    default:
      return -ENOSYS;
  }

  return -ENOSYS;
} /* slink_ioctl() */


/* Multiple process's may open the device */
static int slink_open(struct inode *inode, struct file *filp) {
  unsigned int minor = MINOR(inode->i_rdev);
  struct slink_user *user;
  static int unique_id = 0;

  TRACE("%p, %p", inode, filp);
  if (minor >= SLINK_NO)
    return -ENXIO;

  if (!(slink_table[minor].flags & SLINK_EXIST))
    return -ENXIO;

  if (slink_table[minor].users == 0) {
    if (parport_claim(slink_table[minor].dev) != 0) {
      PRINT(D_ERROR, "Unable to claim device\n");
      return -ENXIO;
    }

    /* Schedule regular reads to decode the interrupt pulses.  Use a
     * tasklet for 2.4.x and the default work queue to 2.6.x kernels
     */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
    slink_table[minor].read_tasklet.next  = NULL;
    slink_table[minor].read_tasklet.state = 0;
    slink_table[minor].read_tasklet.count.counter = 0;
    slink_table[minor].read_tasklet.func = slink_read_tasklet;
    slink_table[minor].read_tasklet.data =(unsigned long)&(slink_table[minor]);
    tasklet_schedule(&(slink_table[minor].read_tasklet));
#else
    INIT_WORK(&(slink_table[minor].read_work),
              slink_read_work, (void *)&(slink_table[minor]));
    schedule_delayed_work(&(slink_table[minor].read_work), 10);
#endif
  }

  /* Allocate a structure for each user using the device */
  KMALLOC(user, sizeof(struct slink_user));
  INIT_LIST_HEAD(&(user->in_queue));
  user->in_queue_depth = 0;
  user->flags = 0;
  user->id = unique_id;
  filp->private_data = (void *)user;
  list_add_tail(&(user->list), &(slink_table[minor].user_queue)); 

#ifdef MOD_INC_USE_COUNT
  MOD_INC_USE_COUNT;
#endif
  slink_table[minor].users++;
  slink_table[minor].flags |= SLINK_BUSY;
  unique_id++;

  return 0;
} /* slink_open() */


static int slink_release(struct inode *inode, struct file *filp) {
  unsigned int minor = MINOR(inode->i_rdev);
  struct list_head *next, *ptr;
  struct slink_packet *packet;
  struct slink_user *user1, *user2;
 
  TRACE("%p, %p", inode, filp); 
#ifdef MOD_DEC_USE_COUNT
  MOD_DEC_USE_COUNT;
#endif
  slink_table[minor].users--;

  user1 = (struct slink_user *)filp->private_data;

  /* Flush the users read queue */
  for (ptr = user1->in_queue.next; ptr != &(user1->in_queue); ptr = next) {
    packet = list_entry(ptr, struct slink_packet, list);
    next = ptr->next;
    KFREE(packet, sizeof(struct slink_packet));
    list_del(ptr);
  }

  /* Remove the user from the device user queue */
  for (ptr = slink_table[minor].user_queue.next;
       ptr != &(slink_table[minor].user_queue); ptr = next) {
    user2 = list_entry(ptr, struct slink_user, list);
    next = ptr->next;

    if (user1->id == user2->id) {
      KFREE(user2, sizeof(struct slink_user));
      list_del(ptr);
    }
  }

  /* Flush the write queue */
  if (slink_table[minor].users == 0) {
    for (ptr = slink_table[minor].out_queue.next;
         ptr != &(slink_table[minor].out_queue); ptr = next) {
      packet = list_entry(ptr, struct slink_packet, list);
      next = ptr->next;
      slink_table[minor].out_queue_depth--;
      KFREE(packet, sizeof(struct slink_packet));
      list_del(ptr);
    }

    parport_release(slink_table[minor].dev);
    slink_table[minor].flags &= ~SLINK_BUSY;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
    cancel_delayed_work(&(slink_table[minor].read_work));
    flush_scheduled_work();
#endif
  }

  return 0;
} /* slink_release() */


/* -------------------- Module Parport Ops  --------------------- */

static int slink_preempt(void *ptr) {
  TRACE("%p", ptr);
  return 0;
} /* slink_preempt() */


static void slink_kick(void *ptr) {
  TRACE("%p", ptr);
} /* slink_kick() */


/* In hindsight it may have been wiser to implement a shift register
 * rather than the circular timestamp buffer.  This may have kept the
 * complexity down in the read_task decode and improved latency
 * to the device. 
 */
static void slink_irq(int irq, void *private, struct pt_regs *regs) {
  struct slink_struct *slink_entry = (struct slink_struct *)private;

//  TRACE("%d, %p, %p", irq, private, regs);  
  if ((slink_entry->flags & SLINK_BUSY) && 
      (!(slink_entry->flags & SLINK_OVERFLOW_IRQ)) &&
      (!(slink_entry->flags & SLINK_UNDERFLOW_IRQ)))  {

    slink_entry->interrupts++;
    if (slink_entry->irqs.size < 0) {
      slink_entry->flags |= SLINK_UNDERFLOW_IRQ;
      slink_entry->irqs.size = 0;
      slink_entry->irqs.head = slink_entry->irqs.buffer;
      slink_entry->irqs.tail = slink_entry->irqs.buffer;
      PRINT(D_WARN, "Interrupt buffer underflow.\n");
    }

    if (slink_entry->irqs.size >= (IRQ_BUFFER_SIZE - sizeof(cycles_t))) {
      slink_entry->flags |= SLINK_OVERFLOW_IRQ;
      PRINT(D_WARN, "Interrupt buffer overflow.\n");
      return;
    }

    SET_INTERVAL(slink_entry->irqs.last, 0);
    *(cycles_t *)slink_entry->irqs.head = slink_entry->irqs.last;
    NEXT_IRQ_HEAD(&(slink_entry->irqs));

    /* Atleast one packet pending schedule a read task */
    if ((slink_entry->irqs.size > (MAX_PACKET_SIZE * 8 * sizeof(cycles_t))) &&
        (!(slink_entry->flags & SLINK_IRQ_FLUSH))) {
      slink_entry->flags |= SLINK_IRQ_FLUSH;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
      tasklet_schedule(&(slink_entry->read_tasklet));
#else
      schedule_delayed_work(&(slink_entry->read_work), 10);
#endif
    }
  }

} /* slink_irq() */


/* -------------------- Module Init/Cleanup  --------------------- */

static int slink_register(int i, struct parport *port) {
  unsigned flag1, flag2;

  TRACE("%d, %p", i, port);
  /* Probe for the slink adaptor by default.  This behavior can be
   * disabled by passing 'slink_probe=0' as a insmod module parmeter.
   */
  if (slink_probe) {
    outb_p(0x00, port->base+1);

    outb_p(0x00, port->base);     /* Set A1 xmit pin to 0, this allows the */
    flag1 = inb_p(port->base+1);  /* A1 line to float high. (0x7f)         */ 

    outb_p(0x01, port->base);     /* Set A1 xmit pin to 1, to hold the A1  */
    flag2 = inb_p(port->base+1);  /* line to ground. (0xbf)                */

    /* If a slink device is present flag1 cannot be the same as flag2 */
    PRINT(D_DEBUG, "Parport %d status:  0x%02X == 0x%02X\n", i, flag1, flag2); 
    if (flag1 == flag2) {
      PRINT(D_INFORM, "Parport %d at 0x%lX irq %d; adaptor NOT found\n",
        i, port->base, port->irq);
      parport_disable_irq(port);
      return 1;
    }
  }
 
  outb_p(0x00, port->base);
  KMALLOC(slink_table[i].irqs.buffer, IRQ_BUFFER_SIZE);

  if (slink_table[i].irqs.buffer == NULL) {
    PRINT(D_ERROR, "Parport %d IRQ buffer not allocated!\n", i);
    return 1;
  }
	  
  if ((slink_table[i].dev = parport_register_device(port,
      "slink", slink_preempt, slink_kick, slink_irq,
      PARPORT_DEV_EXCL, (void *)&slink_table[i])) == NULL) {
    PRINT(D_ERROR, "Parport %d Unable to register device\n", i);
    return 1;
  }

  slink_table[i].flags |= SLINK_EXIST;
  slink_table[i].irqs.head = slink_table[i].irqs.buffer;
  slink_table[i].irqs.tail = slink_table[i].irqs.buffer;
  PRINT(D_INFORM, "Parport %d at 0x%lX irq %d; adaptor found\n",
    i, port->base, port->irq); 
  parport_enable_irq(port);

  return 0;
} /* slink_register() */


static int slink_unregister(int i, struct parport *port) {

  TRACE("%d, %p", i, port);
  if (!(slink_table[i].flags & SLINK_EXIST))
    return 1;

  KFREE(slink_table[i].irqs.buffer, IRQ_BUFFER_SIZE);

  if (slink_table[i].dev != NULL)
    parport_unregister_device(slink_table[i].dev);

  slink_table[i].flags = 0;
  slink_table[i].irqs.buffer = NULL;
  slink_table[i].dev = NULL;

  return 0;
} /* slink_unregister() */


static void slink_attach(struct parport *port) {

  TRACE("%p", port);
  if (slink_size == SLINK_NO) {
    PRINT(D_ERROR, "Exceeded maximum ports (%d)\n", SLINK_NO);
    return;
  }

  if (!slink_register(slink_size, port))
    slink_size++;

  return;
} /* slink_attach() */


static void slink_detach(struct parport *port) {

  TRACE("%p", port);
  if (!slink_unregister(port->portnum, port))
    slink_size--;

  return;
} /* slink_deatch() */


static int __init slink_init(void) {
  int i, ret;

  TRACE("%s", "");
  PRINT(D_INFORM, "v" VERSION "\n");

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
  spin_lock_init(&slink_lock);
#endif

  SET_INTERVAL(slink_time.init, INIT_LENGTH);
  SET_INTERVAL(slink_time.zero, ZERO_LENGTH);
  SET_INTERVAL(slink_time.one, ONE_LENGTH);
  SET_INTERVAL(slink_time.clear, PORT_CLEAR);
  SET_INTERVAL(slink_time.tolerance, PORT_TOLERANCE);
  SET_INTERVAL(slink_time.ir, IR_LENGTH);

  for (i = 0; i < SLINK_NO; i++) {
    slink_table[i].dev         = NULL;
    slink_table[i].flags       = 0;
    slink_table[i].users       = 0;
    slink_table[i].in_packets  = 0;
    slink_table[i].out_packets = 0;
    slink_table[i].corrupted_packets = 0;
    slink_table[i].corrected_packets = 0;
    slink_table[i].collision_packets = 0;
    slink_table[i].interrupts        = 0;
    slink_table[i].out_queue_depth   = 0;
    slink_table[i].irqs.buffer = NULL;
    slink_table[i].irqs.head   = NULL;
    slink_table[i].irqs.tail   = NULL;
    slink_table[i].irqs.size   = 0;
    SET_INTERVAL(slink_table[i].irqs.last, 0);
    INIT_LIST_HEAD(&(slink_table[i].out_queue));
    INIT_LIST_HEAD(&(slink_table[i].user_queue));
    init_waitqueue_head(&(slink_table[i].read_waitq));
    init_waitqueue_head(&(slink_table[i].write_waitq));
    init_MUTEX(&(slink_table[i].user_sem));
    init_MUTEX(&(slink_table[i].read_sem));

    /* Schedule a write for when the port is available */
    slink_table[i].write_tasklet.next  = NULL;
    slink_table[i].write_tasklet.state = 0;
    slink_table[i].write_tasklet.count.counter = 0;
    slink_table[i].write_tasklet.func  = slink_write_tasklet;
    slink_table[i].write_tasklet.data  = (unsigned long)&(slink_table[i]);
  }

  /* Auto-detect available major if not specified */
  if ((ret = register_chrdev(slink_major, "slink", &slink_fops)) < 0) {
    if (slink_major)
      PRINT(D_ERROR, "Major number %i, unavailable.\n", slink_major);
    else
      PRINT(D_ERROR, "No available major numbers.\n");

    return -EIO;
  } else {
    if (ret > 0)
      slink_major = ret;
  }

  /* Register with the parport driver */
  if (parport_register_driver(&slink_driver)) {
    PRINT(D_ERROR, "Unable to register with parport.\n");
    return -EIO;
  }

  /* Create a proc entry */
  create_proc_read_entry("slink", 0, NULL, slink_read_proc, &slink_table);

  if (slink_debug & D_PROC)
    create_proc_read_entry("slink_debug", 0, NULL, 
      slink_debug_read_proc, &slink_table);

  if (slink_size == 0)
    PRINT(D_ERROR, "No parport devices found.\n");

  return 0;
} /* slink_init() */


static int __init slink_init_module(void) {

  if (slink_debug & D_PROC) {
    if ((debug_buffer.buf = vmalloc(DEBUG_BUFFER_SIZE)) == NULL) {
      PRINT(D_ERROR, "Unable to allocate internal debug buffer\n");
      return 0;
    }

    debug_buffer.head = debug_buffer.buf;
    debug_buffer.size = 0;
  }

  TRACE("%s", "");
  if ((slink_major < 0) || (slink_major > 255)) {
    PRINT(D_ERROR, "Specified major %d invalid\n", slink_major);
    return 0;
  }

  return slink_init();
} /* init_module() */


static void slink_cleanup_module(void) {
  int i;

  TRACE("%s", "");
  remove_proc_entry("slink", NULL);

  if (slink_debug & D_PROC)
    remove_proc_entry("slink_debug", NULL);

  parport_unregister_driver(&slink_driver);
  unregister_chrdev(slink_major, "slink");

  for (i = 0; i < SLINK_NO; i++)
    slink_unregister(i, NULL);  

  if (slink_debug & D_PROC) 
    vfree(debug_buffer.buf);

} /* cleanup_module() */


module_init(slink_init_module);
module_exit(slink_cleanup_module);

MODULE_AUTHOR("Brian Behlendorf <behlendorf1@llnl.gov>");
MODULE_DESCRIPTION(PACKAGE " v" VERSION);
MODULE_LICENSE("GPL");

#ifdef MODULE_PARM
MODULE_PARM(slink_major, "1i");
MODULE_PARM(slink_debug, "1i");
MODULE_PARM(slink_probe, "1i");
#else
module_param(slink_major, int, 0);
module_param(slink_debug, int, 0);
module_param(slink_probe, int, 0);
#endif

MODULE_PARM_DESC(slink_major, " Major number to request");
MODULE_PARM_DESC(slink_debug, " Debug mask; default=0x07\n"
  "                             D_WARN   (0x01), D_ERROR (0x02),\n"
  "                             D_INFORM (0x04), D_MEMORY (0x08),\n"
  "                             D_DEBUG  (0x10), D_TRACE (0x20),\n"
  "                             D_PROC   (0x40)");
MODULE_PARM_DESC(slink_probe, " Probe adaptor on module load; default=1");
