/*
 * Slink Deduce - Deduces a device's command set via iteration.
 * Copyright (C) 2003 Brian Behlendorf
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * 2003-Feb-01 Eric Wollesen <ericw@xmtp.net>: Added timeout, device
 * and start code options.  Made code more CD player independent.
 * Comments and cleanup too.
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdlib.h>
#include <slink.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>

/* defines */
#define DEF_DEVICE   0x90 /* CD player 1 */
#define DEF_START    0x00 /* Command to start at */
#define DEF_TIMEOUT  5    /* Time to wait between commands */
#define SLINK_DEVICE "/dev/slink0"
#define SKIP_SIZE    1
 
/* types */
typedef struct Options {
  unsigned int command;
  unsigned int device;
  unsigned int timeout;
} Options;

/* prototypes */
static int parse_args(int argc, char *argv[], Options *options);
static int skip(int key);
static void usage(void);
static int wait_for_rsp(int fd, char command, int timeout);

static int parse_args(int argc, char *argv[], Options *options){

  static const struct option longOptions[] = {
    { "command", optional_argument, 0, 'c' },
    { "device",  optional_argument, 0, 'd' },
    { "timeout", optional_argument, 0, 't' },
    { 0, 0, 0, 0 }
  };
  char *end;
  int option;
  int ret;

  opterr = 0;
  ret = 0;

  while((option = getopt_long(argc, argv, "c:d:t:", longOptions, 
			      (int *)NULL)) != -1){
    switch(option){
    case 'c':
      options->command = strtol(optarg, &end, 16);
      fprintf(stdout, "Start command set to 0x%02X.\n", options->command);
      break;
    case 'd':
      options->device = strtol(optarg, &end, 16);
      fprintf(stdout, "Device set to 0x%02X.\n", options->device);
      break;
    case 't':
      options->timeout = atoi(optarg);
      fprintf(stdout, "Timeout set to %d seconds.\n", options->timeout);
      break;
    default:
      ret = 1;
      usage();
      break;
    }
  }

  return ret;
} /* parse_args() */

static int skip(int key) {
  int i, ignore[SKIP_SIZE] = { 0x2F /* power off */ };

  for (i = 0; i < SKIP_SIZE; i++)
    if (ignore[i] == key)
      return 1;

  return 0;
} /* skip() */

static int wait_for_rsp(int fd, char command, int timeout) {
  char rsp[128];
  int n, size;
  struct timeval tval;
  fd_set rset;

  tval.tv_sec = timeout;
  tval.tv_usec = 0;

  while ((tval.tv_usec + tval.tv_sec * 1000000) > 0) {
    FD_ZERO(&rset);
    FD_SET(fd, &rset);

    if ((n = select(fd + 1, &rset, NULL, NULL, &tval)) < 0) {
      fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      return 1;
    }

    if (n == 0)
      continue;

    if (FD_ISSET(fd, &rset)) {
      if ((size = read(fd, rsp, 128)) < 0) {
        if (errno != ENODATA)
          fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      } else {
        if (command == (char)0x00) {
          fprintf(stdout, "%s\n", rsp);
          return 0;
        }

        if (rsp[1] == command) 
          return 0;
      }
    }
  }

  return 1;
} /* wait_for_rsp() */

static void usage(void){

  fprintf(stdout, "Usage: slink_deduce "
    "[-d device] [-h] [-s start_command] [-t timeout]\n\n");
  fprintf(stdout,
    "  -c, --command COMMAND  starting command command (0x%02X)\n"
    "  -d, --device DEVICE    device code (0x%02X)\n"
    "  -t, --timeout SECONDS  time to wait for a response to a command (%d)\n"
    "  -h, --help             this help message\n",
    DEF_START, DEF_DEVICE, DEF_TIMEOUT);
} /* usage() */

int main(int argc, char *argv[]) {    
  int i, slink;
  char cmd[MAX_PACKET_SIZE], stop[2];
  Options options = { DEF_START, DEF_DEVICE, DEF_TIMEOUT };

  if(parse_args(argc, argv, &options)){
    exit(1);
  }

  if ((slink = open(SLINK_DEVICE, O_RDWR | O_NONBLOCK)) < 0){
    fprintf(stdout, "Could not open device %s.\n", SLINK_DEVICE);
    exit(1);
  }

  fprintf(stdout, "Attempting to deduce available command set.\n\n");

  stop[0] = options.device;
  stop[1] = 0x01; /* command code for stop */

  while (options.command <= 0xFF) {
    memset(cmd, 0, MAX_PACKET_SIZE);
    cmd[0] = options.device;
    cmd[1] = options.command;

    if (skip(options.command)) {
      fprintf(stdout, "--- Skiping 0x%02X 0x%02X... ---\n", 
	      options.device, options.command);
    } else {
      for (i = 2; i <= 6; i++) {
        fprintf(stdout, "--- Sending 0x%02X 0x%02X [%d...] ---\n", 
		options.device, options.command, i-2);

        /* don't go around sending stops after each command unless we
           have reason to believe the device we are deducing is a CD
           player. */
	if(options.device >= 0x90 && options.device <= 0x95){
          do { 
            write(slink, stop, 2);
          } while (wait_for_rsp(slink, (char)0x01, options.timeout));
        }

        ioctl(slink, SLINK_VERBOSE, 1);
        write(slink, cmd, i);
        while (!wait_for_rsp(slink, (char)0x00, options.timeout));
        ioctl(slink, SLINK_VERBOSE, 0);

        fprintf(stdout, "\n");
        fflush(stdout);
      }
    }

    options.command++;
  }

  close(slink);

  return 0;
} /* main() */
