/*
 * Slink Logger - Logs the S-Link data as it comes and goes.
 * Copyright (C) 2003 Brian Behlendorf
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * 2003-Feb-01 Eric Wollesen <ericw@xmtp.net>: Revamped options code.
 * Comments and cleanup.  Also added ability to set timeout to -1 to 
 * allow for infinite running.
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <slink.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>

/* defines */
#define DEF_TIMEOUT    5  /* time in seconds to log data */
#define SLINK_DEVICE   "/dev/slink0"

/* types */
typedef struct Options {
  int timeout;
} Options;

/* prototypes */
static int parse_args(int argc, char *argv[], Options *options);
static void usage(void);

/* parse_args: Get the command line arguments. */
static int parse_args(int argc, char *argv[], Options *options){

  static const struct option longOptions[] = {
    { "timeout", optional_argument, 0, 't' },
    { 0, 0, 0, 0 }
  };
  int option;
  int ret;

  opterr = 0;
  ret = 0;

  while((option = getopt_long(argc, argv, "t:", longOptions, 
			      (int *)NULL)) != -1){
    switch(option){
    case 't':
      options->timeout = atoi(optarg);
      fprintf(stdout, "Timeout set to %d seconds.\n", options->timeout);
      break;
    default:
      ret = 1;
      usage();
      break;
    }
  }

  return(ret);

} /* parse_args() */

/* usage: Print a quick help message. */
static void usage(void){

  fprintf(stdout, "Usage: slink_logger [-t timeout]\n\n");
  fprintf(stdout,
    "  -h, --help             Print this help message\n"
    "  -t, --timeout SECONDS  Time to log, -1 will log indefinitely (%d)\n\n",
    DEF_TIMEOUT); 
} /* usage() */

int main(int argc, char *argv[]) {
  int loop_forever, n, slink, size, seq = 0;
  struct timeval *tval;
  char cmd[1024];
  fd_set rset;
  Options options = { DEF_TIMEOUT };

  if (parse_args(argc, argv, &options))
    exit(1);

  if ((slink = open(SLINK_DEVICE, O_RDWR | O_NONBLOCK)) < 0) {
    fprintf(stdout, "Could not open device.\n");
    exit(1);
  }

  if (options.timeout == -1) {
    loop_forever = 1;
    tval = NULL; /* tells select to block indefinitely */
  } else {
    loop_forever = 0;
    tval = malloc(sizeof(struct timeval));
    tval->tv_sec = options.timeout;
    tval->tv_usec = 0;
  }

  ioctl(slink, SLINK_VERBOSE, 1);
  ioctl(slink, SLINK_ERROR_EXTRA_IRQ, 0);
  fprintf(stdout, "--- log start ---\n");

  while ((loop_forever == 1) || ((tval->tv_usec+tval->tv_sec*1000000) > 0)) {
    FD_ZERO(&rset);
    FD_SET(slink, &rset);

    if ((n = select(slink + 1, &rset, NULL, NULL, tval)) < 0) {
      fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      return 1;
    }

    if (n == 0)
      continue;

    if (FD_ISSET(slink, &rset)) {
      if ((size = read(slink, cmd, 1024)) < 0) {
        if (errno != ENODATA)
          fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      } else {
        fprintf(stdout, "[%d] %s\n", seq, cmd);
        seq++;
      }
    }
  }

  if (loop_forever == 0)
    free(tval);

  close(slink);
  fprintf(stdout, "---- log end ----\n");

  return 0;
} /* main() */
