/* A simple test to estimate the read error rate as perceived
 * by a user space application accessing the slink device.
 * Assuming everything is in harmony the error rate should be
 * zero and the data rate around ~80 bytes/sec (the protocol
 * limits the data rate).  The driver doesn't gaurentee any of
 * this, so use these numbers only as a rule of thumb.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <slink.h>


int query_deck_size(int fd) {
  char rsp[MAX_PACKET_SIZE];
  char cmd1[2] = { 0x90, 0x01 }, cmd2[2] = { 0x90, 0x22 };
  struct timeval tval;
  fd_set rset;
  int n, size;

  write(fd, cmd1, 2);
  write(fd, cmd2, 2); 

  tval.tv_sec = 5;
  tval.tv_usec = 0;

  while ((tval.tv_usec + tval.tv_sec * 1000000) > 0) {
    FD_ZERO(&rset);
    FD_SET(fd, &rset);

    if ((n = select(fd + 1, &rset, NULL, NULL, &tval)) < 0) {
      fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      return -1;
    }

    if (FD_ISSET(fd, &rset)) {
      if ((size = read(fd, rsp, MAX_PACKET_SIZE)) < 0) {
        if (errno != ENODATA)
          fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      } else {
        if (rsp[1] == 0x61)
          return (int)(rsp[2] & 0xFF);
      }
    }
  }

  return -1;
} /* query_deck_size() */


int query_deck_memo(int fd, char *buffer, int deck_size) {
  char *ptr = buffer, cmd[2] = { 0x90, 0x90 };
  struct timeval tval;
  fd_set rset;
  int n, size;

  write(fd, cmd, 2); 

  tval.tv_sec = 90;
  tval.tv_usec = 0;

  while ((tval.tv_usec + tval.tv_sec * 1000000) > 0) {
    FD_ZERO(&rset);
    FD_SET(fd, &rset);

    if ((n = select(fd + 1, &rset, NULL, NULL, &tval)) < 0) {
      fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      return -1;
    }

    if (FD_ISSET(fd, &rset)) {
      if ((size = read(fd, ptr, MAX_PACKET_SIZE)) < 0) {
        if (errno != ENODATA)
          fprintf(stdout, "Error [%d] %s\n", errno, strerror(errno));
      } else {
        if ((((*(char *)(ptr)) == (char)0x9B) ||
            ((*(char *)(ptr)) == (char)0x98)) &&
            (*(char *)(ptr + 2) == (char)deck_size)) {
          ptr += size;
          return (int)(ptr - buffer);
        }

        ptr += size;
      }
    }
  }

  return -1;
} /* query_deck_memo() */


int main(int argc, char *argv[]) {
  char rsp1[8192], rsp2[8192];
  int verbose = 0, deck_size, slink;
  unsigned long i, size1, size2, errors = 0, delay;
  struct timeval start, stop;

  if ((slink = open("/dev/slink0", O_RDWR | O_NONBLOCK)) < 0) {
    fprintf(stdout, "Could not open device.\n");
    return 1;
  }

  ioctl(slink, SLINK_ERROR_EXTRA_IRQ, 0);
  fprintf(stdout,"Validating data integrity (this will take a few minutes)\n");

  if ((deck_size = query_deck_size(slink)) < 0) {
    fprintf(stdout, "Deck 0x90 not responding\n");
    return 1;
  }

  gettimeofday(&start, NULL);
  size1 = query_deck_memo(slink, rsp1, deck_size);
  size2 = query_deck_memo(slink, rsp2, deck_size);
  gettimeofday(&stop, NULL);

  if (size1 != size2) {
    fprintf(stdout, "Error: Packets dropped, check /proc/slink\n");
    return 1;
  }

  if (argc == 2) {
    if (argv[1] == "-v") {
      verbose = 1;
      fprintf(stdout, "\n");
    }
  }

  for (i = 0; i < size1; i++)
    if (rsp1[i] != rsp2[i]) {
      errors++;

      if (verbose)
        fprintf(stdout, "Error at char %lu: %02X != %02X\n",
          i, rsp1[i] & 0xFF, rsp2[i] & 0xFF);
    }

  delay = (double)((stop.tv_sec  * 1000000 + stop.tv_usec) -
	           (start.tv_sec * 1000000 + start.tv_usec));

  fprintf(stdout, "\nError rate: %lu/%lu (~%.3f%%)\n", errors, size1 + size2,
    (double)errors * 100.0 / (double)(size1 + size2));
  fprintf(stdout, "Data rate:  ~%.3f bytes/sec\n",
    (double)(size1 + size2) / (double)delay * 1000000.0);

  close(slink);

  return 0;
} /* main() */
